/*
 Copyright 2006 Foxmarks Inc.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 	http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 foxmarks-manageroverlay.js: hooks the bookmarks manager "import" command
 so we can do do some post processing.

 */



function HandleImport()
{
  var rTarget = RDF.GetResource("NC:BookmarksRoot");
  RDFC.Init(BMDS, rTarget);
  var countBefore = parseInt(BookmarksUtils.getProperty(rTarget, gRDF_NS+"nextVal"));
  
  goDoCommand("cmd_bm_import");
  
  var countAfter = parseInt(BookmarksUtils.getProperty(rTarget, gRDF_NS+"nextVal"));

  // for each child of BookmarksRoot starting at countBefore up to (but not
  // including) countAfter, force the bookmarks service to update the last
  // modified date on each child node so that Foxmarks will recognize them
  // as new items in need of synchronization.

  var rTarget = RDF.GetResource("NC:BookmarksRoot");
  for (var index = countBefore; index < countAfter; index++) {
    var nChildArc = RDFCU.IndexToOrdinalResource(index);
    var rChild    = BMDS.GetTarget(rTarget, nChildArc, true);
    DirtySubTree(rChild);
  }
}

function DirtySubTree(aResource)
{
  var typePred = RDF.GetResource(gRDF_NS + "type");
  var type = BMDS.GetTarget(aResource, typePred, true);
  var isLivemark = (type instanceof Components.interfaces.nsIRDFResource &&
       type.Value == (gNC_NS + "Livemark"));
  var isFolder = (type instanceof Components.interfaces.nsIRDFResource &&
       type.Value == (gNC_NS + "Folder"));
  
  var pred = isLivemark ? RDF.GetResource(gNC_NS + "BookmarkAddDate") :
                          RDF.GetResource(gWEB_NS + "LastModifiedDate");
  var oldval = BMDS.GetTarget(aResource, pred, true);
	var newval = RDF.GetDateLiteral(Date.parse(Date()) * 1000);

  // Oh, how I LOVE the bookmark service. Here's the deal.
  // We want to make sure that this resource is recognized as dirty.
  // For most resource types, that means bumping its last modified date.
  // For livemarks, that means bumping its add date.
  
  // First, we try to find the relevant value (last mod or add date).
  // If we find it, we remove it. That alone will cause the bmsvc to
  // bump the last modified date, adding it back after we've just
  // unasserted it in the non-Livemark case. In the normal case
  // (not a Livemark and there was a lastmod assertion), we're done.
  
  // If there wasn't anything to unassert or it's a livemark that
  // we're dealing with, do the assertion. Again, that will cause bmsvc
  // to bump the last modified date.
  
  //
  if (oldval)
    BMDS.Unassert(aResource, pred, oldval, true);

  if (!oldval || isLivemark)
    BMDS.Assert(aResource, pred, pred, true);
    
  if (isFolder) {          // it's a folder; do the children, too
    RDFC.Init(BMDS, aResource);
    var children = RDFC.GetElements();
    
    while (children.hasMoreElements())
      DirtySubTree(children.getNext());
      
  }
}
